@@ -747,9 +747,13 @@ static int set_network(struct supplicant_task *task,
 						DBUS_TYPE_STRING, &passphrase);
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
 		const char *key_mgmt = "NONE", *index = "0";
+		const char *auth_alg = "OPEN SHARED";
 		connman_dbus_dict_append_variant(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
 
+		connman_dbus_dict_append_variant(&dict, "auth_alg",
+						DBUS_TYPE_STRING, &auth_alg);
+
 		if (passphrase) {
 			int size = strlen(passphrase);
 			if (size == 10 || size == 26) {
@@ -929,7 +933,7 @@ static void extract_ssid(DBusMessageIter *value,
 {
 	DBusMessageIter array;
 	unsigned char *ssid;
-	int ssid_len;
+	int ssid_len, i;
 
 	dbus_message_iter_recurse(value, &array);
 	dbus_message_iter_get_fixed_array(&array, &ssid, &ssid_len);
@@ -948,7 +952,12 @@ static void extract_ssid(DBusMessageIter *value,
 	if (result->name == NULL)
 		return;
 
-	memcpy(result->name, ssid, ssid_len);
+	for (i = 0; i < ssid_len; i++) {
+		if (g_ascii_isprint(ssid[i]))
+			result->name[i] = ssid[i];
+		else
+			result->name[i] = ' ';
+	}
 }
 
 static void extract_wpaie(DBusMessageIter *value,
